import itertools
import os
import numpy
import pandas
import gym
from .tennis import TennisOddsEnv
from ..utils.constants.tennis import CSV_CACHE_PATH, CSV_URL, TOURNAMENTS, SITES, YEARS


class TennisDataMixin(object):
    """Mixin to creating an odds dataframe from www.tennis-data.co.uk"""

    def _create_odds_dataframe(self, tournament, year):
        if tournament not in TOURNAMENTS:
            raise ValueError(f'Tournament {tournament} not supported')
        if year not in YEARS:
            raise ValueError(f'Year {year} not supported')
        if os.path.exists(CSV_CACHE_PATH):
            csvs = [pandas.read_csv(os.path.join(CSV_CACHE_PATH,
                                                 f'{i}',
                                                 f'{tournament}{women}.csv'))
                    for i in range(min(YEARS), year if year != min(YEARS) else year + 1)
                    for women in ['', 'w']]
        else:
            csvs = [pandas.read_csv(CSV_URL.format(tournament=tournament,
                                                   year=i,
                                                   women=women))
                    for i in range(min(YEARS), year if year != min(YEARS) else year + 1)
                    for women in ['', 'w']]
        raw_odds_data = pandas.concat(csvs)
        raw_odds_data['Date'] = pandas.to_datetime(raw_odds_data['Date'], dayfirst=True)
        odds = [''.join(odd) for odd in itertools.product(SITES, ['W', 'L'])
                if ''.join(odd) in raw_odds_data.columns]
        odds_dataframe = raw_odds_data[['Winner', 'Loser', 'Date'] + odds].copy()
        odds_dataframe.rename({'Winner': 'winner', 'Loser': 'loser', 'Date': 'date'},
                              axis='columns', inplace=True)
        odds_dataframe['result'] = 0
        for odd in ['W', 'L']:
            odds_columns = [column for column in odds_dataframe.columns if column.endswith(odd)]
            odds_dataframe[f'Max{odd}'] = odds_dataframe[odds_columns].max(axis='columns')
            odds_dataframe[f'Avg{odd}'] = odds_dataframe[odds_columns].mean(axis='columns')
            odds_dataframe[f'Min{odd}'] = odds_dataframe[odds_columns].min(axis='columns')
            odds_dataframe[f'Median{odd}'] = odds_dataframe[odds_columns].median(axis='columns')
        odds_dataframe
        return odds_dataframe


class TennisDataDailyEnv(TennisDataMixin, TennisOddsEnv):
    sport = 'Tennis, with odds from www.tennis-data.co.uk'
    versionadded = '0.8.0'

    ENV_COLUMNS = ['winner', 'loser', 'date', 'result']
    ODDS_COLUMNS = ['win', 'lose']

    def __init__(self, tournament='ausopen', year=2010, columns='max', extra=False,
                 optimize='reward', *args, **kwargs):
        """Initializes a new environment

        Parameters
        ----------
        tournament: str, defaults to "ausopen"
            The name of the tournament.
        year: int, defaults to 2010
            Year of the tournament.
        columns: {"max", "avg"}, defaults to "max"
            Which columns to use for the odds data. "max" uses the maximum value
            for each odd from all sites, "avg" uses the average value.
        extra: bool, default to False
            Use extra odds for the observation space.
        optimize: {"balance", "reward"}, default to "balance"
            Which type of optimization to use.
        """
        odds_dataframe = self._create_odds_dataframe(tournament, year)
        odds_dataframe.rename({f'{columns.title()}W': 'win',
                               f'{columns.title()}L': 'lose'}, axis='columns', inplace=True)
        super().__init__(odds_dataframe[self.ENV_COLUMNS + self.ODDS_COLUMNS],
                         *args, **kwargs)
        self._extra_odds = odds_dataframe
        self._extra = extra
        self._optimize = optimize
        if self._extra:
            self.observation_space = gym.spaces.Box(low=0., high='Inf',
                                                    shape=(self.observation_space.shape[0],
                                                           self._extra_odds.shape[1] - len(self.ENV_COLUMNS)))

    def step(self, action):
        obs, reward, done, info = super().step(action)
        if self._extra:
            obs = self.get_extra_odds()
        if self._optimize == 'balance':
            if not info['legal_bet']:
                self.balance -= self.starting_bank * 1e-6
            reward = self.balance - self.starting_bank
        else:
            if not info['legal_bet']:
                reward = -(self.starting_bank * 1e-6)
        return obs, reward, done, info

    def get_extra_odds(self):
        extra_odds = numpy.zeros([*self.observation_space.shape])
        current_odds = self._extra_odds.iloc[self._get_current_index()].drop(self.ENV_COLUMNS, axis='columns').values
        extra_odds[numpy.arange(current_odds.shape[0])] = current_odds
        return extra_odds


class TennisDataDailyPercentageEnv(TennisDataDailyEnv):
    versionadded = '0.8.0'



import numpy
from gym import spaces
from .base import BaseOddsEnv


class BasePercentageOddsEnv(BaseOddsEnv):
    """Base class for sports betting environments with non fixed bet size.

    Creates an OpenAI Gym environment that supports betting a non fixed amount
    on a single outcome for a single game.

    The main difference between the BaseOddsEnv is that the action space is defined
    differently (to accommodate that non fixed bet size).

    .. versionadded:: 0.2.0
    .. versionchanged:: 0.4.5
        Name changed to "BasePercentageOddsEnv"

    Parameters
    ----------
    action_space : gym.spaces.Box of shape (N,), where N is the number of possible
        outcomes for the game.
        The indexes are the percentage of the current balance to place on matching
        outcome, so that action[i] is the bet percentage for outcome[i].

        .. versionchanged:: 0.5.0
            Change action space so that each outcome has it's own independent bet
            percentage.
        .. versionchanged:: 0.6.0
            Change action space bounds to [-1, 1] and rescale the action back
            inside the step method.
        .. versionchanged:: 0.7.0
            Reduce dimesionality of action space, deduce the action implicitly
            from the given percentages

        The rescaling an action :math:`(a, p_0, ..., p_{N-1})\\in\\text{action_space}, -1 \\leq a, p_0, ..., p_{N-1} \\leq 1`:

        .. math::
            \\begin{cases}
                a' = \\lfloor (a + 1) * (2^{N-1}) \\rfloor\\\\
                p_i' = |p_i|
            \\end{cases}

    """

    def __init__(self, odds, odds_column_names, results=None):
        super().__init__(odds, odds_column_names, results)
        self.bet_size_matrix = None
        self.action_space = spaces.Box(low=numpy.array([-1] * odds.shape[1]),
                                       high=numpy.array([1.] * odds.shape[1]))

    def step(self, action):
        form_binary_repr = numpy.where(action != 0, 1, 0)
        form = form_binary_repr.dot(1 << numpy.arange(form_binary_repr.size)[::-1])
        self.bet_size_matrix = self._rescale_matrix(action) * self.balance
        return super().step(form)



import pandas
from .meta import MetaEnvBuilder


class ThreeWaySoccerOddsEnv(metaclass=MetaEnvBuilder):
    sport = '3-way soccer'
    versionadded = '0.1.0'
    odds_column_names = ['home', 'draw', 'away']

    def __init__(self, soccer_bets_dataframe):
        """Initializes a new environment.

        Parameters
        ----------
        soccer_bets_dataframe: dataframe of shape (n_games, n_odds + 3), n_odds == 3
            A list of games, with their betting odds, results and the teams playing.

            .. warning::

                Please make sure that the odds columns are named "home, draw, away",
                that the results column is named "result" and that the team names
                columns are named "home_team, away_team" respectively.
        """
        odds = soccer_bets_dataframe[self.odds_columns]
        results = soccer_bets_dataframe['result'] if soccer_bets_dataframe['result'].notna().all() else None
        if not self.daily_env:
            odds = odds.values
            results = results.values
        self.teams = soccer_bets_dataframe[['home_team', 'away_team']]
        super().__init__(odds, self.odds_column_names, results)

    def render(self, mode='human'):
        """Outputs the current team names, balance and step.

        Returns
        -------
        msg : str
            A string with the current team names, balance and step.
        """
        index = self._get_current_index()
        teams = self.teams.iloc[index]
        teams = teams.itertuples() if isinstance(teams, pandas.DataFrame) else [teams]
        teams_str = ', '.join(['Home Team {} VS Away Team {}'.format(row.home_team, row.away_team)
                               for row in teams])
        teams_str = teams_str + "."
        print(teams_str)
        super().render(mode)


class ThreeWaySoccerPercentageOddsEnv(ThreeWaySoccerOddsEnv):
    versionadded = '0.2.0'


class ThreeWaySoccerDailyOddsEnv(ThreeWaySoccerOddsEnv):
    versionadded = '0.5.0'


class ThreeWaySoccerDailyPercentageOddsEnv(ThreeWaySoccerOddsEnv):
    versionadded = '0.5.0'



import os


CSV_URL = 'http://www.football-data.co.uk/mmz4281/{start}{end}/{country}{league}.csv'

COUNTRIES = {'England': 'E', 'Scotland': 'SC', 'Germany': 'D', 'Italy': 'I', 'Spain': 'SP', 'France': 'F',
             'Netherlands': 'N', 'Belgium': 'B', 'Portugal': 'P', 'Turkey': 'T', 'Greece': 'G'}

LEAGUES = {'England': {'Premier League': 0, 'Championship': 1, 'League 1': 2, 'League 2': 3, 'Conference': 'C'},
           'Scotland': {'Premier League': 0, 'Division 1': 1, 'Division 2': 2, 'Division 3': 3},
           'Germany': {'Bundesliga 1': 1, 'Bundesliga 2': 2},
           'Italy': {'Serie A': 1, 'Serie B': 2},
           'Spain': {'Primera Division': 1, 'Segunda Division': 2},
           'France': {'Ligue 1': 1, 'Ligue 2': 2},
           'Netherlands': {'Eredivise': 1},
           'Belgium': {'Jupiler League': 1},
           'Portugal': {'Liga 1': 1},
           'Turkey': {'Super Lig': 1},
           'Greece': {'Super League': 1}}

CSV_CACHE_PATH = os.path.abspath(os.path.join(os.path.abspath(__file__), '..', 'csv_cache', 'football'))

SITES = ["B365", "BS", "BW", "GB", "IW", "LB", "PS", "SO", "SB", "SJ", "SY", "VC", "WH", "P"]

YEARS = range(10, 20)


def url_kwargs():
    for country in LEAGUES:
        for league in LEAGUES[country].values():
            yield {'country': COUNTRIES[country], 'league': league}



import gym
import numpy
import numexpr
from itertools import compress
from pandas import DataFrame
from tabulate import tabulate


class ActionsDict(dict):
    """A custom dictionary implementation to simplfy getting verbose actions"""

    def __init__(self, odds_column_names, *args, **kwargs):
        self.odds_column_names = odds_column_names
        super().__init__(self, *args, **kwargs)

    def __missing__(self, key):
        self[key] = [list(compress(self.odds_column_names,
                                   [int(bit) for bit in numpy.binary_repr(key, width=len(self.odds_column_names))]))]
        return self[key]


class BaseOddsEnv(gym.Env):
    """Base class for sports betting environments.

    Creates an OpenAI Gym environment that supports betting a fixed amount (1)
    on a single outcome for a single game.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    observation_space : gym.spaces.Box
        The observation space for the environment.
        The observation space shape is (1, N) where N is the number of possible
        outcomes for the game.

        .. versionchanged:: 0.3.0
            Changed definition of space

    action_space : gym.spaces.Discrete
        The action space for the environment.
        The action space is a single number in [0, 2 ** N), representing
        on what outcomes to place a bet by conversion to a binary representation.

        .. versionchanged:: 0.3.0
            Changed definition of space

    balance : float
        The current balance of the environment.

    starting_bank : int, default=10
        The starting bank / balance for the environment.
    """

    metadata = {'render.modes': ['human']}
    HEADERS = ['Current Step', 'Odds', 'Verbose Action', 'Action', 'Balance', 'Reward', 'Results', 'Done']

    def __init__(self, odds, odds_column_names, results=None, starting_bank=10):
        """Initializes a new environment

        Parameters
        ----------
        odds: dataframe of shape (n_games, n_odds)
            A list of games, with their betting odds.
        odds_column_names: list of str
            A list of column names with length == n_odds.
        results: list of int, default=None
            A list of the results, where results[i] is the outcome of odds[i].
        """

        super().__init__()
        self._odds = odds.copy()
        self._results = results
        self._odds_columns_names = odds_column_names
        self._verbose_actions = ActionsDict(self._odds_columns_names)
        self.observation_space = gym.spaces.Box(low=1., high=float('Inf'), shape=(1, odds.shape[1]))
        self.action_space = gym.spaces.Discrete(2 ** odds.shape[1])
        self.balance = self.starting_bank = starting_bank
        self.current_step = 0
        self.bet_size_matrix = numpy.ones(shape=self.observation_space.shape)

    def _get_current_index(self):
        return self.current_step % self._odds.shape[0]

    def get_odds(self):
        """Returns the odds for the current step.

        Returns
        -------
        odds : numpy.ndarray of shape (1, n_odds)
            The odds for the current step.
        """
        return DataFrame([self._odds[self.current_step]]).values

    def get_bet(self, action):
        """Returns the betting matrix for the action provided.

        Parameters
        ----------
        action : int
            An action provided by the agent.

        Returns
        -------
        bet : array of shape (1, n_odds)
            The betting matrix, where each outcome specified in the action
            has a value of 1 and 0 otherwise.
        """
        return numpy.array([[bit for bit in numpy.binary_repr(action, width=self._odds.shape[1])]]).astype(int)

    def step(self, action):
        """Run one timestep of the environment's dynamics. When end of episode is reached,
        you are responsible for calling reset() to reset this environment's state.

        Accepts an action and returns a tuple (observation, reward, done, info).

        Parameters
        ----------
        action : int
            An action provided by the agent.

        Returns
        -------
        observation : dataframe of shape (1, n_odds)
            The agent's observation of the current environment
        reward : float
            The amount of reward returned after previous action
        done : bool
            Whether the episode has ended, in which case further step() calls will return undefined results
        info : dict
            Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning)

        """
        odds = self.get_odds()
        reward = 0
        done = False
        info = self.create_info(action)
        if self.balance < 1:  # no more money :-(
            done = True
        else:
            bet = self.get_bet(action)
            results = self.get_results()
            if self.legal_bet(bet):  # making sure agent has enough money for the bet
                reward = self.get_reward(bet, odds, results)
                self.balance += reward
                info.update(legal_bet=True)
            else:
                reward = -(bet * self.bet_size_matrix).sum()
            info.update(results=results.argmax())
            info.update(reward=reward)
            self.current_step += 1
            if self.finish():
                done = True
                odds = numpy.ones(shape=self.observation_space.shape)
            else:
                odds = self.get_odds()
        info.update(done=done)
        return odds, reward, done, info

    def get_reward(self, bet, odds, results):
        """ Calculates the reward

        Parameters
        ----------
        bet : array of shape (1, n_odds)
        odds: dataframe of shape (1, n_odds)
            A games with its betting odds.
        results : array of shape (1, n_odds)

        Returns
        -------
        reward : float
            The amount of reward returned after previous action
        """
        bet_size_matrix = self.bet_size_matrix  # noqa: F841
        reward = numexpr.evaluate('sum(bet * bet_size_matrix * results * odds)')
        expense = numexpr.evaluate('sum(bet * bet_size_matrix)')
        return reward - expense

    def reset(self):
        """Resets the state of the environment and returns an initial observation.

        Returns
        -------
        observation : dataframe of shape (1, n_odds)
            the initial observation.
        """
        self.balance = self.starting_bank
        self.current_step = 0
        return self.get_odds()

    def render(self, mode='human'):
        """Outputs the current balance and the current step.

        Returns
        -------
        msg : str
            A string with the current balance and the current step.
        """
        print('Current balance at step {}: {}'.format(self.current_step, self.balance))

    def finish(self):
        """Checks if the episode has reached an end.

        The episode has reached an end if there are no more games to bet.

        Returns
        -------
        finish : bool
            True if the current_step is equal to n_games, False otherwise
        """
        return self.current_step == self._odds.shape[0]  # no more games left to bet

    def get_results(self):
        """Returns the results matrix for the current step.

        Returns
        -------
        result : array of shape (1, n_odds)
            The result matrix, where the index of the outcome that happened
            value is 1 and the rest of the indexes values are 0.
        """
        result = numpy.zeros(shape=self.observation_space.shape)
        result[numpy.arange(result.shape[0], dtype=numpy.int32),
               numpy.array([self._results[self.current_step]], dtype=numpy.int32)] = 1
        return result

    def legal_bet(self, bet):
        """Checks if the bet is legal.

        Checks that the bet does not exceed the current balance.

        Parameters
        ----------
        bet : array of shape (1, n_odds)
            The bet to check.

        Returns
        -------
        legal : bool
            True if the bet is legal, False otherwise.
        """
        return (bet * self.bet_size_matrix).sum() <= self.balance

    def create_info(self, action):
        """Creates the info dictionary for the given action.

        The info dictionary holds the following information:
            * the verbose action
            * the current step
            * the balance at the start of the current step
            * the relevant odds for the current step
            * the bet size for a single outcome

        Parameters
        ----------
        action : int
            An action provided by the agent.

        Returns
        -------
        info : dict
            The info dictionary.
        """
        return {'current_step': self.current_step, 'odds': self.get_odds(), 'verbose_action': self._verbose_actions[action],
                'action': action, 'balance': self.balance, 'reward': 0,
                'legal_bet': False, 'results': None, 'done': False}

    def pretty_print_info(self, info):
        values = [info[key.replace(' ', '_').lower()] for key in self.HEADERS]
        print('\n' + tabulate([values], headers=self.HEADERS, tablefmt='orgtbl'))

    def _rescale_form(self, form):
        if form == 1:
            form -= numpy.finfo(numpy.float64).eps
        return numpy.floor((form + 1) * (2 ** (self._odds.shape[1] - 1))).astype(int)

    def _rescale_matrix(self, bet_size_matrix):
        return numpy.abs(bet_size_matrix)



import datetime
import itertools
import os
import numpy
import pandas
import gym
from .soccer import ThreeWaySoccerOddsEnv
from ..utils.constants.football import CSV_CACHE_PATH, CSV_URL, COUNTRIES, LEAGUES, SITES


class FootballDataMixin(object):
    """Mixin to creating an odds dataframe from www.football-data.co.uk"""

    def _create_odds_dataframe(self, country, league, start, end):
        country = country.title()
        league = league.title()
        if country not in COUNTRIES:
            raise ValueError(f'Country {country} not supported')
        if league not in LEAGUES[country]:
            raise ValueError(f'League {league} not supported')
        if start < 2010:
            raise ValueError(f'Start year must be greater than 2010, start year is {start}')
        if start > end:
            raise ValueError('Start year can not be greater than end year')
        if end > datetime.datetime.today().year:
            raise ValueError('End year can not be greater than current year')
        if os.path.exists(CSV_CACHE_PATH):
            csvs = [pandas.read_csv(os.path.join(CSV_CACHE_PATH,
                                                 f'{str(i).zfill(2)}{str(i + 1).zfill(2)}',
                                                 f'{COUNTRIES[country]}{LEAGUES[country][league]}.csv'))
                    for i in range(int(str(start)[-2:]), int(str(end)[-2:]))]
            raw_odds_data = pandas.concat(csvs)
        else:
            raw_odds_data = pandas.concat([pandas.read_csv(CSV_URL.format(country=COUNTRIES[country],
                                                                          league=LEAGUES[country][league],
                                                                          start=str(i).zfill(2),
                                                                          end=str(i + 1).zfill(2)))
                                           for i in range(int(str(start)[-2:]), int(str(end)[-2:]))])
        raw_odds_data['Date'] = pandas.to_datetime(raw_odds_data['Date'], dayfirst=True)
        odds = [''.join(odd) for odd in itertools.product(SITES, ['H', 'D', 'A'])
                if ''.join(odd) in raw_odds_data.columns]
        odds_dataframe = raw_odds_data[['HomeTeam', 'AwayTeam', 'Date'] + odds].copy()
        odds_dataframe.rename({'HomeTeam': 'home_team', 'AwayTeam': 'away_team', 'Date': 'date'},
                              axis='columns', inplace=True)
        odds_dataframe['result'] = raw_odds_data['FTR'].map({'H': 0, 'A': 2, 'D': 1})
        odds_dataframe.dropna(subset=['result'], inplace=True)
        odds_dataframe['result'] = odds_dataframe['result'].astype(int)
        for odd in ['H', 'D', 'A']:
            odds_columns = [column for column in odds_dataframe.columns if column.endswith(odd)]
            odds_dataframe[f'Max{odd}'] = odds_dataframe[odds_columns].max(axis='columns')
            odds_dataframe[f'Avg{odd}'] = odds_dataframe[odds_columns].mean(axis='columns')
            odds_dataframe[f'Min{odd}'] = odds_dataframe[odds_columns].min(axis='columns')
            odds_dataframe[f'Median{odd}'] = odds_dataframe[odds_columns].median(axis='columns')
        odds_dataframe
        return odds_dataframe


class FootballDataDailyEnv(FootballDataMixin, ThreeWaySoccerOddsEnv):
    sport = '3-way soccer, with odds from www.football-data.co.uk'
    versionadded = '0.7.1'

    ENV_COLUMNS = ['home_team', 'away_team', 'date', 'result']
    ODDS_COLUMNS = ['home', 'draw', 'away']

    def __init__(self, country='England', league='Premier League', start=2010, end=2011, columns='max', extra=False,
                 optimize='reward', *args, **kwargs):
        """Initializes a new environment

        Parameters
        ----------
        country: str, defaults to "England"
            The name of the country in which the league is playing.
        league: str, defaults to "Premier League"
            The name of the league.
        start: int, defaults to 2010
            Start year of the league to use.
        end: int, defaults to 2011
            End year of the league to use.
        columns: {"max", "avg"}, defaults to "max"
            Which columns to use for the odds data. "max" uses the maximum value
            for each odd from all sites, "avg" uses the average value.
        extra: bool, default to False
            Use extra odds for the observation space.
        optimize: {"balance", "reward"}, default to "balance"
            Which type of optimization to use.
        """
        odds_dataframe = self._create_odds_dataframe(country, league, start, end)
        odds_dataframe.rename({f'{columns.title()}H': 'home',
                               f'{columns.title()}D': 'draw',
                               f'{columns.title()}A': 'away'}, axis='columns', inplace=True)
        super().__init__(odds_dataframe[self.ENV_COLUMNS + self.ODDS_COLUMNS],
                         *args, **kwargs)
        self._extra_odds = odds_dataframe
        self._extra = extra
        self._optimize = optimize
        if self._extra:
            self.observation_space = gym.spaces.Box(low=0., high='Inf',
                                                    shape=(self.observation_space.shape[0],
                                                           self._extra_odds.shape[1] - len(self.ENV_COLUMNS)))

    def step(self, action):
        obs, reward, done, info = super().step(action)
        if self._extra:
            obs = self.get_extra_odds()
        if self._optimize == 'balance':
            if not info['legal_bet']:
                self.balance -= self.starting_bank * 1e-6
            reward = self.balance - self.starting_bank
        else:
            if not info['legal_bet']:
                reward = -(self.starting_bank * 1e-6)
        return obs, reward, done, info

    def get_extra_odds(self):
        extra_odds = numpy.zeros([*self.observation_space.shape])
        current_odds = self._extra_odds.iloc[self._get_current_index()].drop(self.ENV_COLUMNS, axis='columns').values
        extra_odds[numpy.arange(current_odds.shape[0])] = current_odds
        return extra_odds


class FootballDataDailyPercentageEnv(FootballDataDailyEnv):
    versionadded = '0.6.2'



import pandas
from .meta import MetaEnvBuilder


class TennisOddsEnv(metaclass=MetaEnvBuilder):
    sport = 'tennis'
    versionadded = '0.8.0'
    odds_column_names = ['win', 'lose']

    def __init__(self, tennis_bets_dataframe):
        """Initializes a new environment.

        Parameters
        ----------
        tennis_bets_dataframe: dataframe of shape (n_matches, n_odds + 3), n_odds == 2
            A list of matches, with their betting odds, results and the players playing.

            .. warning::

                Please make sure that the odds columns are named "win, lose",
                that the results column is named "result" and that the player names
                columns are named "winner, loser" respectively.
        """
        odds = tennis_bets_dataframe[self.odds_columns]
        results = tennis_bets_dataframe['result'] if tennis_bets_dataframe['result'].notna().all() else None
        if not self.daily_env:
            odds = odds.values
            results = results.values
        self.players = tennis_bets_dataframe[['winner', 'loser']]
        super().__init__(odds, self.odds_column_names, results)

    def render(self, mode='human'):
        """Outputs the current player names, balance and step.

        Returns
        -------
        msg : str
            A string with the current player names, balance and step.
        """
        index = self._get_current_index()
        players = self.players.iloc[index]
        players = players.itertuples() if isinstance(players, pandas.DataFrame) else [players]
        players_str = ', '.join(['Player {} VS Player {}'.format(row.winner, row.loser)
                                 for row in players])
        players_str = players_str + "."
        print(players_str)
        super().render(mode)


class TennisPercentageOddsEnv(TennisOddsEnv):
    pass


class TennisDailyOddsEnv(TennisOddsEnv):
    pass


class TennisDailyPercentageOddsEnv(TennisOddsEnv):
    pass



import gym
import numpy
import numexpr
from pandas import DataFrame
from .base import BaseOddsEnv


class DailyOddsEnv(BaseOddsEnv):
    """Base class for sports betting environments multiple games with a fixed
    bet size.

    Creates an OpenAI Gym environment that supports betting a fixed amount (1)
    on a single outcome but for multiple games.

    .. versionadded:: 0.3.0

    Parameters
    ----------
    observation_space : gym.spaces.Box
        The observation space for the environment.
        The observation space shape is (M, N) where M is the maximum number of games in a single
        day and N is the number of possible outcomes for the game.

    action_space : gym.spaces.Box
        The action space for the environment.
        The action space shape is (M,), a list of numbers in [0, 2 ** N), representing
        on what outcomes to place a bet by conversion to a binary representation,
        where actions[i] is the action for odds[i].

        .. versionchanged:: 0.6.0
            Change action space bounds to [-1, 1] and rescale the action back
            inside the step method.
            For explanation on how the rescaling works, see :py:class:`~oddsgym.envs.base_percentage.BasePercentageOddsEnv`

    balance : float
        The current balance of the environment.

    starting_bank : int, default=10
        The starting bank / balance for the environment.

    days : array, dtype=np.datetime64
        Sorted array of the days.

    """

    HEADERS = ['Date', 'Current Step', 'Odds', 'Verbose Action', 'Action', 'Balance', 'Reward', 'Results', 'Done']

    def __init__(self, odds, odds_column_names, results=None, max_number_of_games='auto'):
        """Initializes a new environment.

        We initialize the days array (because this environment doesn't iterate
        the games a single game at a time, but by a group of games that happened
        in the same date) and calculate the day with the most games in it so
        that the observation & action spaces will be defined correctly.

        Parameters
        ----------
        odds: dataframe of shape (n_games, n_odds + 1)
            A list of games, with their betting odds and the date in which
            the game occurs.

            .. warning::
                Please note that the environment expects the date column
                in the dataframe to be named 'date' (lower cased).

        odds_column_names: list of str
            A list of column names with length == n_odds.

        results: list of int, default=None
            A list of the results, where results[i] is the outcome of odds[i].

        max_number_of_games: int or str, default='auto'
            The maximum number of games in a single name that the environment
            will support. The value 'auto' will calculate this value from the
            odds dataframe.
        """
        super().__init__(odds.drop('date', 'columns'), odds_column_names, results)
        self._odds_with_dates = odds.copy()
        self.days = odds['date'].unique()
        self.days.sort()
        self._max_number_of_games = None
        if max_number_of_games == 'auto':
            self.max_number_of_games = odds.set_index('date').groupby(by='date').size().max()
        elif isinstance(max_number_of_games, int) and max_number_of_games > 0:
            self.max_number_of_games = max_number_of_games
        else:
            raise ValueError("Invalid value for max_number_of_games: {}.\nPass 'auto' for automatic "
                             "calculation of the maximum number of games"
                             ", or an integer higher than 0".format(max_number_of_games))
        self.observation_space = gym.spaces.Box(low=0., high=float('Inf'),
                                                shape=(self.max_number_of_games, self._odds.shape[1]))
        self.action_space = gym.spaces.Box(low=-1,
                                           high=1,
                                           shape=(self.max_number_of_games,))
        self.bet_size_matrix = numpy.ones(shape=self.observation_space.shape)

    def _get_current_index(self):
        current_day = self.days[self.current_step]
        return self._odds_with_dates[self._odds_with_dates['date'] == current_day].index

    def get_odds(self):
        """Returns the odds for the current step.

        Returns
        -------
        odds : numpy.ndarray of shape (max_games, n_odds)
            The odds for the current step.
            If in the current step there were less then max_games, the odds
            dataframe is appended with (max_games - current_games) zeroed rows
            (rows with only 0).
        """
        current_odds = self._odds.iloc[self._get_current_index()]
        filler_odds = DataFrame(numpy.zeros(numpy.array([self.max_number_of_games, self._odds.shape[1]]) -
                                            numpy.array([current_odds.shape[0], 0])),
                                columns=self._odds_columns_names)
        return current_odds.append(filler_odds, ignore_index=True).values

    def get_bet(self, action):
        """Returns the betting matrix for the action provided.

        Parameters
        ----------
        action : array of shape (max_games, )
            An action provided by the agent.

        Returns
        -------
        bet : array of shape (max_games, n_odds)
            The betting matrix, where for each row, each outcome specified in
            action[row] has a value of 1 and 0 otherwise.
        """
        full_actions = numpy.zeros(shape=(self.max_number_of_games, self._odds.shape[1]))
        actions = numpy.concatenate([super(DailyOddsEnv, self).get_bet(numpy.floor(part_action).astype(int))
                                     for part_action in action])
        full_actions[numpy.arange(actions.shape[0])] = actions
        return full_actions

    def get_results(self):
        """Returns the results matrix for the current step.

        Returns
        -------
        results : array of shape (max_games, n_odds)
            The results matrix, where for each row (single game),
            the index of the outcome that happened value is 1 and the rest of
            the indexes values are 0.
            If in the current step there were less then max_games, the result
            matrix is appended with (max_games - current_games) zeroed rows
            (rows with only 0).
        """
        index = self._get_current_index()
        current_results = self._results.iloc[index]
        results = numpy.zeros(shape=(current_results.shape[0], self._odds.shape[1]))
        results[numpy.arange(results.shape[0]), current_results.values] = 1
        filler_results = numpy.zeros(numpy.array([self.max_number_of_games, self._odds.shape[1]]) -
                                     numpy.array([current_results.shape[0], 0]))
        return numpy.concatenate([results, filler_results])

    def get_reward(self, bet, odds, results):
        """ Calculates the reward, while taking to account invalid bets

        Parameters
        ----------
        bet : array of shape (n_games, n_odds)
        odds: dataframe of shape (n_games, n_odds)
            A list of games, with their betting odds.
        results : array of shape (max_games, n_odds)

        Returns
        -------
        reward : float
            The amount of reward returned after previous action
        """
        used_results = numpy.ones_like(results)
        zero_rows_count = numpy.sum(~results.any(1))
        if zero_rows_count > 0:
            used_results[-zero_rows_count:, :] = 0
        bet_size_matrix = self.bet_size_matrix  # noqa: F841
        reward = numexpr.evaluate('sum(bet * bet_size_matrix * results * odds)')
        expense = numexpr.evaluate('sum(bet * used_results * bet_size_matrix)')
        return reward - expense

    def legal_bet(self, bet):
        """Checks if the bet is legal, while taking to account invalid bets.

        Checks that the bet does not exceed the current balance.

        Parameters
        ----------
        bet : array of shape (n_games, n_odds)
            The bet to check.

        Returns
        -------
        legal : bool
            True if the bet is legal, False otherwise.
        """
        results = self.get_results()
        used_results = numpy.ones_like(results)
        zero_rows_count = numpy.sum(~results.any(1))
        if zero_rows_count > 0:
            used_results[-zero_rows_count:, :] = 0
        return (bet * used_results * self.bet_size_matrix).sum() <= self.balance

    def finish(self):
        """Checks if the episode has reached an end.

        The episode has reached an end if there are no more days left.

        Returns
        -------
        finish : bool
            True if the current_step is equal to len(self.days), False otherwise
        """
        return self.current_step == self.days.shape[0]  # no more days left to bet

    def create_info(self, action):
        """Creates the info dictionary for the given action.

        The info dictionary holds the following information:
            * the verbose actions
            * the current step
            * the balance at the start of the current step
            * the relevant odds for the current step
            * the bet size for a single outcome

        Parameters
        ----------
        action : array of shape (max_games,)
            An action provided by the agent.

        Returns
        -------
        info : dict
            The info dictionary.
        """
        return {'date': self.days[self.current_step], 'current_step': self.current_step,
                'odds': self.get_odds(),
                'verbose_action': [self._verbose_actions[act] for act in numpy.floor(action).astype(int)],
                'action': action,
                'balance': self.balance, 'reward': 0,
                'legal_bet': False, 'results': None, 'done': False}

    def step(self, action):
        return super().step(numpy.array([self._rescale_form(form) for form in action]))


class DailyPercentageOddsEnv(DailyOddsEnv):
    """Base class for sports betting environments multiple games with a non fixed
    bet size.

    Creates an OpenAI Gym environment that supports betting a non fixed amount
    on a single outcome but for multiple games.

    .. versionadded:: 0.5.0

    Parameters
    ----------
    action_space: gym.spaces.Box of shape (n_games * n_odds,)
        A vector with the action and the percentage for each outcome, so that:

        .. math::

            a[i] = \\begin{cases}
                \\text{the action} & i\\mod\\text{(n_odds + 1)}\\equiv 0\\\\
                \\text{percentage of outcome j} & i\\mod\\text{(n_odds + 1)}\\equiv j
            \\end{cases}

        For the game :math:`\\lfloor \\frac{i}{\\text{n_games}} \\rfloor`.

        .. versionchanged:: 0.6.0
            Change action space bounds to [-1, 1] and rescale the action back
            inside the step method.
            For explanation on how the rescaling works, see :py:class:`~oddsgym.envs.base_percentage.BasePercentageOddsEnv`
        .. versionchanged:: 0.7.0
            Reduce dimesionality of action space by deducing action from
            betting percentages

    """

    def __init__(self, odds, odds_column_names, results=None, *args, **kwargs):
        super().__init__(odds, odds_column_names, results, *args, **kwargs)
        lower_bound = [[-1] * self._odds.shape[1] for i in numpy.arange(self.max_number_of_games)]
        upper_bound = [[1] * self._odds.shape[1] for i in numpy.arange(self.max_number_of_games)]
        vector_size = self.max_number_of_games * self._odds.shape[1]
        self.action_space = gym.spaces.Box(low=numpy.array(lower_bound).reshape(vector_size),
                                           high=numpy.array(upper_bound).reshape(vector_size))

    def step(self, action):
        full_action = numpy.zeros(self.max_number_of_games * (self._odds.shape[1]))
        full_action[numpy.arange(action.shape[0])] = action
        full_action = full_action.reshape(self.max_number_of_games, (self._odds.shape[1]))

        current_bet_size_matrix = self._rescale_matrix(full_action) * self.balance
        full_bet_size_matrix = numpy.zeros(shape=(self.max_number_of_games, self._odds.shape[1]))
        full_bet_size_matrix[numpy.arange(current_bet_size_matrix.shape[0])] = current_bet_size_matrix

        self.bet_size_matrix = full_bet_size_matrix

        form_binary_repr = numpy.where(self.bet_size_matrix != 0, 1, 0)
        explicit_forms = form_binary_repr.dot(1 << numpy.arange(form_binary_repr.shape[-1] - 1, -1, -1))
        forms = numpy.array([(form / (2 ** (self._odds.shape[1] - 1)) - 1) for form in explicit_forms])

        return super().step(forms)



import gym

from .base import BaseOddsEnv
from .base_percentage import BasePercentageOddsEnv
from .daily_bets import DailyOddsEnv, DailyPercentageOddsEnv
from .soccer import ThreeWaySoccerOddsEnv, ThreeWaySoccerPercentageOddsEnv
from .soccer import ThreeWaySoccerDailyOddsEnv, ThreeWaySoccerDailyPercentageOddsEnv
from .footballdata import FootballDataDailyEnv, FootballDataDailyPercentageEnv
from .tennisdata import TennisDataDailyEnv, TennisDataDailyPercentageEnv


__all__ = ['BaseOddsEnv', 'BasePercentageOddsEnv', 'DailyOddsEnv', 'DailyPercentageOddsEnv', 'ThreeWaySoccerOddsEnv',
           'ThreeWaySoccerPercentageOddsEnv', 'ThreeWaySoccerDailyOddsEnv', 'ThreeWaySoccerDailyPercentageOddsEnv',
           'FootballDataDailyEnv', 'FootballDataDailyPercentageEnv',
           'TennisDataDailyEnv', 'TennisDataDailyPercentageEnv']

# register the *-data.co.uk environments, so they would be ready for usage when installing with pip
gym.register(id='FootballDataDaily-v0',
             entry_point='oddsgym.envs:FootballDataDailyEnv',
             max_episode_steps=365)
gym.register(id='FootballDataDailyPercent-v0',
             entry_point='oddsgym.envs:FootballDataDailyPercentageEnv',
             max_episode_steps=365)
gym.register(id='TennisDataDaily-v0',
             entry_point='oddsgym.envs:TennisDataDailyEnv',
             max_episode_steps=365)
gym.register(id='TennisDataDailyPercent-v0',
             entry_point='oddsgym.envs:TennisDataDailyPercentageEnv',
             max_episode_steps=365)

# register the *-data.co.uk environments in ray, only if ray is installed
try:
    from ray.tune.registry import register_env
except ImportError:  # pragma: no cover
    pass
else:
    register_env('FootballDataDaily-ray-v0', lambda env_config: FootballDataDailyEnv(**env_config))
    register_env('FootballDataDailyPercent-ray-v0', lambda env_config: FootballDataDailyPercentageEnv(**env_config))
    register_env('TennisDataDaily-ray-v0', lambda env_config: TennisDataDailyEnv(**env_config))
    register_env('TennisDataDailyPercent-ray-v0', lambda env_config: TennisDataDailyPercentageEnv(**env_config))



import os
import asyncio
from contextlib import closing
import aiohttp
import typer
from oddsgym.utils.constants import SupportedSport


async def download_file(session, url_template, **kwargs):
    url = url_template.format(**kwargs)
    while True:
        try:
            async with session.get(url) as response:
                assert response.status == 200, url
                return (kwargs.get("country", kwargs.get("tournament")),
                        kwargs.get("league", kwargs.get("women")),
                        await response.read())
        except AssertionError:  # pragma: no cover
            asyncio.sleep(1)


async def download_multiple(fast, sport):
    url_kwargs = list(sport.url_kwargs)
    years = sport.years
    if fast:
        years = years[-1:]
    with typer.progressbar(years) as years_progress:
        for year in years_progress:
            year_csv_path = os.path.join(sport.csv_cache_url,
                                         sport.csv_url.split('/')[-2].format(start=year, end=year + 1,
                                                                             year=year, women=''))
            if not os.path.exists(year_csv_path):
                os.mkdir(year_csv_path)
            async with aiohttp.ClientSession() as session:
                download_futures = [download_file(session, sport.csv_url, start=year,
                                                  end=year + 1, year=year, **kwargs)
                                    for kwargs in url_kwargs]
                for download_future in asyncio.as_completed(download_futures):
                    prefix, suffix, result = await download_future
                    with open(os.path.join(year_csv_path, f'{prefix}{suffix}.csv'), 'w') as csv_file:
                        csv_file.write(result.decode('latin-1'))


def create_csv_cache(fast: bool = typer.Option(False, "--fast", "-f", help="Run only on last seasons / years"),
                     sport: SupportedSport = typer.Option(SupportedSport.football, "--sport", "-s",
                                                          help="Which sport to run", case_sensitive=False)):
    if not os.path.exists(sport.csv_cache_url):
        os.makedirs(sport.csv_cache_url)
    asyncio.set_event_loop(asyncio.new_event_loop())
    with closing(asyncio.get_event_loop()) as loop:
        loop.run_until_complete(download_multiple(fast, sport))


def main():  # pragma: no cover
    typer.run(create_csv_cache)



from .base import BaseOddsEnv
from .base_percentage import BasePercentageOddsEnv
from .daily_bets import DailyOddsEnv, DailyPercentageOddsEnv


class MetaEnvBuilder(type):
    def __new__(cls, name, bases, attr):
        def safe_get(attribute):
            if attribute in attr:
                return attr[attribute]
            for base in bases:
                if getattr(base, attribute, None):
                    return getattr(base, attribute)

        new_bases = list(bases)
        percentage_env, daily_env = 'Percentage' in name, 'Daily' in name
        docstring = (f'Environment for {safe_get("sport")} betting{", grouped by date," if daily_env else ""}'
                     f' with a {"non fixed" if percentage_env else "fixed"} bet size.\n\n    '
                     f'.. versionadded:: {safe_get("versionadded")}\n    ')
        if percentage_env and daily_env:
            new_bases += [DailyPercentageOddsEnv]
        elif daily_env:
            new_bases += [DailyOddsEnv]
        elif percentage_env:
            new_bases += [BasePercentageOddsEnv]
        else:
            new_bases += [BaseOddsEnv]
        attr['percentage_env'] = percentage_env
        attr['daily_env'] = daily_env
        attr['odds_columns'] = safe_get('odds_column_names') + ['date'] * daily_env
        attr['odds_column_names'] = safe_get('odds_column_names')
        attr['__doc__'] = docstring
        return super(MetaEnvBuilder, cls).__new__(cls, name, tuple(new_bases), attr)



from enum import Enum
from . import football
from . import tennis


SPORT_CSV_URLS = {"football": football.CSV_URL,
                  "tennis": tennis.CSV_URL}

SPORT_CSV_CACHE_PATHS = {"football": football.CSV_CACHE_PATH,
                         "tennis": tennis.CSV_CACHE_PATH}

SPORT_YEARS = {"football": football.YEARS,
               "tennis": tennis.YEARS}

SPORT_KWARGS = {"football": football.url_kwargs,
                "tennis": tennis.url_kwargs}


class SupportedSport(str, Enum):
    football = "football"
    tennis = "tennis"

    @property
    def csv_url(self):
        return SPORT_CSV_URLS[self.value]

    @property
    def csv_cache_url(self):
        return SPORT_CSV_CACHE_PATHS[self.value]

    @property
    def years(self):
        return SPORT_YEARS[self.value]

    @property
    def url_kwargs(self):
        return SPORT_KWARGS[self.value]()



import os
from itertools import product


CSV_URL = 'http://www.tennis-data.co.uk/{year}{women}/{tournament}.csv'

TOURNAMENTS = ['ausopen', 'frenchopen', 'usopen']

CSV_CACHE_PATH = os.path.abspath(os.path.join(os.path.abspath(__file__), '..', 'csv_cache', 'tennis'))

SITES = ["B365", "B&W", "CB", "EX", "IW", "LB", "PS", "SB", "SJ", "GB", "UB"]

YEARS = range(2010, 2020)


def url_kwargs():
    for tournament, women in product(TOURNAMENTS, ['', 'w']):
        yield {'tournament': tournament, 'women': women}
